<?php
require_once ('boConfig.php');

/*
for MySQL performacne analysis you may want to use the following setup:
1) configuration of the MySQL logging
mysql
SET GLOBAL slow_query_log = 'ON';
SET GLOBAL slow_query_log_file = '/var/log/mysql/localhost-slow.log';
SET GLOBAL log_queries_not_using_indexes = 'ON';
SET SESSION long_query_time = 1;
exit

2) analysis of the MySQL logging
mysqldumpslow -t 200 -s at /var/log/mysql/localhost-slow.log
*/

define("DEBUG_LOGGING", false); // true for logging enabled, false for no logging, this is independent from the MySQL logging described above

class boDb {
    function __construct($text = "") {
        // if enabled, create the logging file
        if (DEBUG_LOGGING) {
            $this->logFile = fopen("boDb.log", "a") or die("Unable to open file!\n");
            $this->log("__construct('$text')");
        }

        $servername = DB_HOST;
        $username = DB_USER;
        $password = DB_PASS;
        $dbname = DB_NAME;
        $dbport = DB_PORT;

        // create connection
        $this->conn = mysqli_connect($servername, $username, $password, $dbname, $dbport) or die("Connection failed: " . mysqli_connect_error() . "\n");
    }

    function __destruct() {
        // close connection
        mysqli_close($this->conn);
        $this->conn = null;

        // close the logging file
        if (DEBUG_LOGGING) {
            $this->log("__destruct()");
            fclose($this->logFile);
        }
    }

    private function log($data) {
        if (DEBUG_LOGGING) {
            $text = "";
            if (is_array($data)) {
                $hasFunction = false;
                if (array_key_exists("function", $data)) {
                    $hasFunction = true;
                    $text .= $data["function"] . "(";
                    unset($data["function"]);
                }
                foreach ($data as $key => $value) {
                    if (is_array($value)) {
                        $text .= "$key={" ;
                            foreach($value as $k => $v) {
                                if (is_array($v)) {
                                    $text .= "$k=array(), ";
                                } else {
                                    $text .= "$k='$v', ";
                                }
                            }
                            $text = rtrim($text);
                            $text = rtrim($text, ",");
                        $text .= "}, ";
                    } else {
                        $text .= "$key='$value', ";
                    }
                }
                $text = rtrim($text);
                $text = rtrim($text, ",");
                if ($hasFunction) {
                    $text .= ")";
                }
            } else {
                $text = $data;
            }
            fwrite($this->logFile, "$text\n");
        }
    }

    // Assume all data older than one entire day plus the hours of the current day is stable.
    // Example:
    //  $maxTime = 1529785200000000000 = Saturday, 23. June 2018 20:20:00
    //  return(1529625600000000000 = Friday, 22. June 2018 00:00:00)
    protected function _calcStableMaxTime($maxTime) {
        $secondsOfDay = 24 * 60 * 60;
        $maxTime = intval(substr($maxTime, 0, -9));
        $maxTime -= $secondsOfDay;
        $maxTime = intval($maxTime / $secondsOfDay) * $secondsOfDay;
        return("$maxTime" . "000000000");
    }

    private function _generiqSqlRequest($sql) {
        if (DEBUG_LOGGING) {
            $t1 = microtime(true);
        }
        $data = null;
        $sql = str_replace(" WHERE 1 AND ", " WHERE ", $sql);
        $result = $this->conn->query($sql);
        if ("boolean" == gettype($result)) {
            $data = $result;
        } else {
            if (1 == $result->num_rows) {
                $data = $result->fetch_assoc();
            } else if (1 < $result->num_rows) {
                $data = array();
                // read the data of each row
                while($row = $result->fetch_assoc()) {
                    $data[] = $row;
                }
            }
        }
        if (DEBUG_LOGGING) {
            $t = microtime(true) - $t1;
            if ($t > 1) {
                $t = number_format($t, 3) . " sec";
            } else if (($t * 1000) > 1) {
                $t = number_format($t * 1000, 3) . " ms";
            } else {
                $t = number_format($t * 1000000, 3) . " us";
            }
            $this->log("$sql [$t]");
        }
        return($data);
    }

    protected function _getSumOfStrokesPerChannel($minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("_getSumOfStrokesPerChannel(minTime=$minTime, maxTime=$maxTime)");
        }
        $sql = "SELECT count(*) as total";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $sql .= ", sum((channels >> $c) & 1) as sum$c";
        }
        $sql .= " FROM `strokes`";
        $sql .= " WHERE channels";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    public function getSumOfStrokesPerChannel($minTime = null, $maxTime = null, $useHistoricalData = false) {
        if (DEBUG_LOGGING) {
            $this->log("getSumOfStrokesPerChannel(minTime=$minTime, maxTime=$maxTime, useHistoricalData=$useHistoricalData)");
        }
        $attribute = array();
        if ($useHistoricalData) {
            $entity = "getSumOfStrokesPerChannel";
            $attribute = array(
                "minTime" => $minTime,
                "maxTime" => $maxTime
            );
            $historicalData = $this->_getHistoricalData($entity, $attribute);
            if (!is_null($historicalData)) {
                $historicalData = json_decode($historicalData["value"], true);
                return($historicalData); // historical data is present, return it
            }
        }
        $sql = "SELECT count(*) as total";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $sql .= ", sum((channels >> $c) & 1) as sum$c";
        }
        $sql .= " FROM `strokes`";
        $sql .= " WHERE channels";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        $result = $this->_generiqSqlRequest($sql); // retreive current data
        if ($useHistoricalData) {
            $this->_setHistoricalData($entity, $attribute, $result); // store current data as historical data
        }
        return($result);
    }

    private function _getSumOfUniqueStrokesForSingleChannel($channel, $minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("_getSumOfUniqueStrokesForSingleChannel(channel=$channel, minTime=$minTime, maxTime=$maxTime)");
        }
        $channelMask = 1 << $channel;
        $sql = "SELECT count(*) as sum$channel";
        $sql .= " FROM `strokes`";
        $sql .= " WHERE channels = $channelMask";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    public function getSumOfUniqueStrokesPerChannel($minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("getSumOfUniqueStrokesPerChannel(minTime=$minTime, maxTime=$maxTime)");
        }
        $result = array();
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $result += $this->_getSumOfUniqueStrokesForSingleChannel($c, $minTime, $maxTime);
        }
        return $result;
    }

    protected function _getSumOfSignalsPerChannel($minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("_getSumOfSignalsPerChannel(minTime=$minTime, maxTime=$maxTime)");
        }
        $sql = "SELECT count(*) as total";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $sql .= ", sum((channels >> $c) & 1) as sum$c";
        }
        $sql .= " FROM `signals`";
        $sql .= " WHERE 1";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns >= $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns < $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    public function getSumOfSignalsPerChannel($minTime = null, $maxTime = null, $useHistoricalData = false) {
        if (DEBUG_LOGGING) {
            $this->log("getSumOfSignalsPerChannel(minTime=$minTime, maxTime=$maxTime, useHistoricalData=$useHistoricalData)");
        }
        $attribute = array();
        if ($useHistoricalData) {
            $entity = "getSumOfSignalsPerChannel";
            $attribute = array(
                "minTime" => $minTime,
                "maxTime" => $maxTime
            );
            $historicalData = $this->_getHistoricalData($entity, $attribute);
            if (!is_null($historicalData)) {
                $historicalData = json_decode($historicalData["value"], true);
                return($historicalData); // historical data is present, return it
            }
        }
        $sql = "SELECT count(*) as total";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $sql .= ", sum((channels >> $c) & 1) as sum$c";
        }
        $sql .= " FROM `signals`";
        $sql .= " WHERE 1";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns >= $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns < $maxTime";
        }
        $result = $this->_generiqSqlRequest($sql); // retreive current data
        if ($useHistoricalData) {
            $this->_setHistoricalData($entity, $attribute, $result); // store current data as historical data
        }
        return($result);
    }

    public function getSumOfUniqueSignalsPerChannel($minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("getSumOfUniqueSignalsPerChannel(minTime=$minTime, maxTime=$maxTime)");
        }
        $sql = "SELECT ";
        $where = " WHERE channels IN (";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            $channelMask = 1 << $c;
            if (0 < $i) {
                $sql .= ", ";
                $where .= ", ";
            }
            $sql .= "sum(CASE WHEN channels=$channelMask THEN 1 ELSE 0 END) as sum$c";
            $where .= "$channelMask";
        }
        $where .= ")";
        $sql .= " FROM `signals`";
        if (is_null($minTime) | is_null($maxTime) | ((intval(substr($maxTime, 0, -9)) - intval(substr($minTime, 0, -9))) > (24 * 60 * 60))) {
            // use the index only if the time is more than one day
            $sql .= " USE INDEX (channels)";
        }
        $sql .= $where;
        if (!is_null($minTime)) {
            $sql .= " AND time_ns >= $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns < $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    protected function _getMinMaxStrokeDistance($channel, $minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("_getMinMaxStrokeDistance(channel=$channel, minTime=$minTime, maxTime=$maxTime)");
        }
        $channelMask = 1 << $channel;
        $sql  = "SELECT min(distance) as min, max(distance) as max FROM `strokes` WHERE channels & $channelMask";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    public function getMinMaxStrokeDistance($channel, $minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("getMinMaxStrokeDistance(channel=$channel, minTime=$minTime, maxTime=$maxTime)");
        }
        $useHistoricalData = is_null($minTime) && !is_null($maxTime); // only this case is implemented to use historical data
        if ($useHistoricalData) {
            $stableMaxTime = $this->_calcStableMaxTime($maxTime);
            // get historical data
            $entity = "getMinMaxStrokeDistance";
            $attribute = array(
                "channel" => $channel
                );
            $historicalData = $this->_getHistoricalData($entity, $attribute);
            if (is_array($historicalData)) {
                // historical data is present
                $id = $historicalData["id"];
                $historicalData = json_decode($historicalData["value"], true);

                if ($historicalData["stableMaxTime"] < $stableMaxTime) {
                    // the historical data is outdated
                    // query the data that has been added between $historicalData["stableMaxTime"] and $stableMaxTime
                    $tmpMinTime = $historicalData["stableMaxTime"];
                    $tmpMaxTime = $stableMaxTime;
                    $resultNew = $this->_getMinMaxStrokeDistance($channel, $tmpMinTime, $tmpMaxTime);
                    if ($resultNew["min"] < $historicalData["min"]) {
                        $historicalData["min"] = $resultNew["min"];
                    }
                    if ($resultNew["max"] > $historicalData["max"]) {
                        $historicalData["max"] = $resultNew["max"];
                    }
                    $historicalData["stableMaxTime"] = $stableMaxTime;
                    // update the historcal data in the database
                    $this->_setHistoricalData($entity, $attribute, $historicalData, $id);
                }
            } else {
                // historical data has to be created
                $tmpMinTime = null;
                $tmpMaxTime = $stableMaxTime;
                $historicalData = $this->_getMinMaxStrokeDistance($channel, $tmpMinTime, $tmpMaxTime);
                $historicalData["stableMaxTime"] = $stableMaxTime;
                // update the historcal data in the database
                $this->_setHistoricalData($entity, $attribute, $historicalData);
            }
            assert($historicalData["stableMaxTime"] == $stableMaxTime);
            // the historical data can be used
            // query the data that has been added after $stableMaxTime
            $tmpMinTime = $stableMaxTime;
            $tmpMaxTime = $maxTime;
            $resultNew = $this->_getMinMaxStrokeDistance($channel, $tmpMinTime, $tmpMaxTime);
            if ($historicalData["min"] < $resultNew["min"]) {
                $resultNew["min"] = $historicalData["min"];
            }
            if ($historicalData["max"] > $resultNew["max"]) {
                $resultNew["max"] = $historicalData["max"];
            }
            $result = $resultNew;
        } else {
            $result = $this->_getMinMaxStrokeDistance($channel, $minTime, $maxTime);
        }
        return($result);
    }

    protected function _getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("_getSumOfStrokesPerDistancePerChannel(minDistance=$minDistance, maxDistance=$maxDistance, minTime=$minTime, maxTime=$maxTime)");
        }
        $sql = "SELECT";
        for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
            $c = CHANNEL_NUMBER[$i];
            if (0 < $i) $sql .= ",";
            $sql .= " sum((channels >> $c) & 1) as sum$c";
        }
        $sql .= " FROM `strokes`";
        $sql .= " WHERE channels";
        $sql .= " AND distance >= $minDistance";
        $sql .= " AND distance < $maxDistance";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        return($this->_generiqSqlRequest($sql));
    }

    public function getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $minTime = null, $maxTime = null) {
        if (DEBUG_LOGGING) {
            $this->log("getSumOfStrokesPerDistancePerChannel(minDistance=$minDistance, maxDistance=$maxDistance, minTime=$minTime, maxTime=$maxTime)");
        }
        $useHistoricalData = is_null($minTime) && !is_null($maxTime); // only this case is implemented to use historical data
        if ($useHistoricalData) {
            $stableMaxTime = $this->_calcStableMaxTime($maxTime);

            // get historical data
            $entity = "getSumOfStrokesPerDistancePerChannel";
            $attribute = array(
                "minDistance" => $minDistance,
                "maxDistance" => $maxDistance
                );
            $historicalData = $this->_getHistoricalData($entity, $attribute);
            if (is_array($historicalData)) {
                // historical data is present
                $id = $historicalData["id"];
                $historicalData = json_decode($historicalData["value"], true);

                if ($historicalData["stableMaxTime"] < $stableMaxTime) {
                    // the historical data is outdated
                    // query the data that has been added between $historicalData["stableMaxTime"] and $stableMaxTime
                    $tmpMinTime = $historicalData["stableMaxTime"];
                    $tmpMaxTime = $stableMaxTime;
                    $resultNew = $this->_getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $tmpMinTime, $tmpMaxTime);
                    for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
                        $c = CHANNEL_NUMBER[$i];
                        $historicalData["sum$c"] += $resultNew["sum$c"];
                    }
                    $historicalData["stableMaxTime"] = $stableMaxTime;
                    // update the historcal data in the database
                    $this->_setHistoricalData($entity, $attribute, $historicalData, $id);
                }
            } else {
                // historical data has to be created
                $tmpMinTime = null;
                $tmpMaxTime = $stableMaxTime;
                $historicalData = $this->_getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $tmpMinTime, $tmpMaxTime);
                $historicalData["stableMaxTime"] = $stableMaxTime;
                // update the historcal data in the database
                $this->_setHistoricalData($entity, $attribute, $historicalData);
            }
            assert($historicalData["stableMaxTime"] == $stableMaxTime);
            // the historical data can be used
            // query the data that has been added after $stableMaxTime
            $tmpMinTime = $stableMaxTime;
            $tmpMaxTime = $maxTime;
            $resultNew = $this->_getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $tmpMinTime, $tmpMaxTime);
            for($i=0; $i<count(CHANNEL_NUMBER); $i++) {
                $c = CHANNEL_NUMBER[$i];
                $resultNew["sum$c"] += $historicalData["sum$c"];
            }
            $result = $resultNew;
        } else {
            $result = ($this->_getSumOfStrokesPerDistancePerChannel($minDistance, $maxDistance, $minTime, $maxTime));
        }
        return($result);
    }

    function getMaxIdOfStrokes() {
        if (DEBUG_LOGGING) {
            $this->log("getMaxIdOfStrokes()");
        }
        $sql = "SELECT max(id) as maxId FROM `strokes` WHERE channels";
        return($this->_generiqSqlRequest($sql)["maxId"]);
    }

    private function _getStrokePositionForChannelsFromIdList($idList, $numMostDistant) {
        if (DEBUG_LOGGING) {
            $this->log(array("function" => "_getStrokePositionForChannelsFromIdList", "idList" => $idList, "numMostDistant" => $numMostDistant));
        }
        $sql = "SELECT id, lat_stroke as lat, lon_stroke as lon, channels, distance";
        $sql .= " FROM `strokes`";
        $sql .= " WHERE id IN (";
        for ($i=0; $i<count($idList); $i++) {
            if (0 < $i) {
                $sql .= ", ";
            }
            $sql .= $idList[$i];
        }
        $sql .= ")";
        $sql .= " ORDER BY distance DESC";
        $sql .= " LIMIT $numMostDistant";
        return($this->_generiqSqlRequest($sql));
    }

    public function getStrokePositionForChannels($minTime = null, $maxTime = null, $minAzimuth = null, $maxAzimuth = null, $numMostDistant = null) {
        if (DEBUG_LOGGING) {
            $this->log("getStrokePositionForChannels(minTime=$minTime, maxTime=$maxTime, minAzimuth=$minAzimuth, maxAzimuth=$maxAzimuth, numMostDistant=$numMostDistant)");
        }
        $useHistoricalData = !is_null($numMostDistant);
        if ($useHistoricalData) {
            $newMaxId = $this->getMaxIdOfStrokes(); // maximum ID at the start of the queries
            $sql = "SELECT lat_stroke as lat, lon_stroke as lon, channels, id, distance";
            $sql .= " FROM `strokes`";
        } else {
            $sql = "SELECT lat_stroke as lat, lon_stroke as lon, channels";
            $sql .= " FROM `strokes`";
        }
        $sql .= " WHERE channels";
        if (!is_null($minTime)) {
            $sql .= " AND time_ns_stroke > $minTime";
        }
        if (!is_null($maxTime)) {
            $sql .= " AND time_ns_stroke <= $maxTime";
        }
        if (!is_null($minAzimuth)) {
            $sql .= " AND azimuth > $minAzimuth";
        }
        if (!is_null($maxAzimuth)) {
            $sql .= " AND azimuth <= $maxAzimuth";
        }
        if ($useHistoricalData) {
            $entity = "getStrokePositionForChannels";
            $attribute = array(
                "minAzimuth" => $minAzimuth,
                "maxAzimuth" => $maxAzimuth
                );
            $historicalData = $this->_getHistoricalData($entity, $attribute);
            $id = null;
            if (!is_null($historicalData)) {
                $id = $historicalData["id"];
                $historicalData = json_decode($historicalData["value"], true);
                $minDistanceOfNumMostDistant = $historicalData["minDistanceOfNumMostDistant"];
                $maxId = $historicalData["maxId"];
                $idList = $historicalData["idList"];
                $sql .= " AND id > $maxId";
                $sql .= " AND distance >= $minDistanceOfNumMostDistant";
                $historicalResult = $this->_getStrokePositionForChannelsFromIdList($idList, $numMostDistant);
                if (is_array($historicalResult)) { // if there is a result ...
                    if (!array_key_exists(0, $historicalResult)) { // and it is a single entry ...
                        $historicalResult = array($historicalResult); // convert it to an array (with a single entry)
                    }
                }
            } else {
                $historicalData = array();
                $historicalResult = array();
            }
            $sql .= " ORDER BY distance DESC";
        }
        if (!is_null($numMostDistant)) {
            $sql .= " LIMIT $numMostDistant";
        }
        $result = $this->_generiqSqlRequest($sql);
        if ($useHistoricalData) {
            if (is_array($result)) { // if there is a result ...
                if (!array_key_exists(0, $result)) { // and it is a single entry ...
                    $result = array($result); // convert it to an array (with a single entry)
                }
                $mergedResult = array();
                $h = 0;
                $r = 0;
                while ((($h + $r) < (count($historicalResult) + count($result))) && (($h + $r) < $numMostDistant)) {
                    if (array_key_exists($h, $historicalResult)) {
                        // entry in the historicalResult exists
                        if (array_key_exists($r, $result)) {
                            // entry in the result exists
                            if ($historicalResult[$h]["distance"] > $result[$r]["distance"]) {
                                // distance of historicalResult is bigger
                                $mergedResult[] = $historicalResult[$h];
                                $h += 1;
                            } else {
                                // distance of result is equal or bigger
                                $mergedResult[] = $result[$r];
                                $r += 1;
                            }
                        } else {
                            // no more entries in the result
                            $mergedResult[] = $historicalResult[$h];
                            $h += 1;
                        }
                    } else {
                        // no more entries in the historicalResult
                        $mergedResult[] = $result[$r];
                        $r += 1;
                    }
                }
            } else {
                $mergedResult = $historicalResult;
            }

            // update the historical data
            $updatedHistoricalData = array();

            // update the maximum index queried
            $updatedHistoricalData["maxId"] = $newMaxId;

            // update the list of ids
            $idList = array();
            for ($i=0; $i<count($mergedResult); $i++) {
                $idList[] = $mergedResult[$i]["id"];
            }
            $updatedHistoricalData["idList"] = $idList;

            // update the distance of the nearest of the N most distant
            if (count($mergedResult)) {
                $indexOfNearest = count($mergedResult) - 1; // due to " ORDER BY distance DESC" the last entry is the nearest
                $minDistanceOfNumMostDistant = floatval($mergedResult[$indexOfNearest]["distance"]);
                $updatedHistoricalData["minDistanceOfNumMostDistant"] = $minDistanceOfNumMostDistant;

                if ($updatedHistoricalData != $historicalData) {
                    $this->_setHistoricalData($entity, $attribute, $updatedHistoricalData, $id);
                }

                return($mergedResult);
            } else {
                return(null);
            }
        }
        return($result);
    }

    public function getStationPosition() {
        if (DEBUG_LOGGING) {
            $this->log("getStationPosition()");
        }
        $sql = "SELECT AVG(`lat_station`) as lat, AVG(`lon_station`) as lon ";
        $sql .= "FROM (SELECT lat_station, lon_station FROM `strokes` ORDER by `id` DESC LIMIT 100) AS position";
        return($this->_generiqSqlRequest($sql));
    }

    private function _setHistoricalData($entity, $attribute, $value, $id = null) {
        if (DEBUG_LOGGING) {
            $this->log(array("function" => "_setHistoricalData", "entity" => $entity, "attribute" => $attribute, "value" => $value, "id" => $id));
        }
        // convert and escape the parameters
        $entity = $this->conn->real_escape_string("$entity");
        $attribute = $this->conn->real_escape_string(json_encode($attribute));
        $value = $this->conn->real_escape_string(json_encode($value));

        if (is_null($id)) {
            // check whether the entity/attribute combination is already known
            $sql = "SELECT * FROM historicalData WHERE entity = '$entity' AND attribute = '$attribute'";
            $result = $this->_generiqSqlRequest($sql);
            if (!is_null($result)) {
                $id = $result["id"];
            }
        }
        if (is_null($id)) {
            // unknown, thus insert
            $sql = "INSERT INTO historicalData (entity, attribute, value) VALUES ('$entity', '$attribute', '$value');";
            if (true != $this->_generiqSqlRequest($sql)) {
                throw new Exception("Can't insert historicalData");
            }
        } else {
            // known, thus update
            $sql = "UPDATE historicalData SET value = '$value' WHERE id = $id;";
            if (true != $this->_generiqSqlRequest($sql)) {
                throw new Exception("Can't update historicalData");
            }
        }
    }

    private function _getHistoricalData($entity, $attribute) {
        if (DEBUG_LOGGING) {
            $this->log(array("function" => "_getHistoricalData", "entity" => $entity, "attribute" => $attribute));
        }
        // convert and escape the parameters
        $entity = $this->conn->real_escape_string("$entity");
        $attribute = $this->conn->real_escape_string(json_encode($attribute));

        $sql = "SELECT id, value FROM historicalData WHERE entity = '$entity' AND attribute = '$attribute'";
        return($this->_generiqSqlRequest($sql));
    }
}
?>