#!/usr/bin/python
from __future__ import print_function
from boConfig import BO_USER, BO_PASS, BO_STATION, BO_REGION, BOOK_KEEPING_PATH, LOGGING_PATH, LOGLEVEL
import time
import calendar
import subprocess
import json
import boDb
import sys
import os
import logging

desiredExecutionTime = 4 * 60 # note: cron starts this job every 10 minutes, thus some reserve should be there
secStep = 10 * 60 # 10 minutes in seconds
bookKeepingFile = "boGetSignalsFromBlitzortung.txt"
logFile = "boGetSignalsFromBlitzortung.log"

url = "http://{0}:{1}@data.lightningmaps.org/Protected/Signals/{2}/%Y/%m/%d/%Y%m%d_%H%M_s{2}.json"

new = 0
done = 0
dropped = 0

def configLogger():
    global logger
    global logFile
    format = "%(asctime)-15s %(levelname)s:%(name)s: %(message)s"

    path = LOGGING_PATH
    if ("" == path):
        path = "./"
    if (os.path.exists(path) and os.path.isdir(path)):
        if ('/' == path[-1]):
            filename = path + logFile
        else:
            filename = path + '/' + logFile
    else:
        print("ERROR: LOGGING_PATH = '{}' does not exist".format(LOGGING_PATH))
        sys.exit(1)

    logging.basicConfig(filename = filename, format = format)
    logger = logging.getLogger(__name__)
    logger.setLevel(LOGLEVEL)
    logger.setLevel(logging.INFO) # overrule LOGLEVEL
    boDb.configLogger()

def calcRequiredDownloads():
    """1 download to be done every 10 minutes from 3 servers
    server may not be reachable => retrials may be required
    server may also not have data => no endless retries
    file for bookkeeping
    first line: start and end time/date in human readable format
    second line: end time/date in human readable format
    following lines: downloads to be done from start to end (inclduing borders))
    succeeded lines are deleted
    missing lines for newer dates are to be added
    """
    global secStep
    global bookKeepingFile
    global new

    fileDateTimeFormat = "%Y.%m.%d %H:%M"
    firstLine = True

    urls = []
    secEnd = (int(time.time() / secStep) * secStep) - secStep # multiple of secStep, older than now
    try:
        with open(bookKeepingFile, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if (line):
                    urls.append(line)
                if (firstLine):
                    firstLine = False
                    line = line.split('-')
                    tInitialStart = time.strptime(line[0].strip(), fileDateTimeFormat)
                    secInitialStart = int(calendar.timegm(tInitialStart) / secStep) * secStep
                    tInitialEnd = time.strptime(line[1].strip(), fileDateTimeFormat)
                    secInitialEnd = int(calendar.timegm(tInitialEnd) / secStep) * secStep
                    secStart = secInitialEnd + secStep
    except:
        secInitialStart = secEnd - 3 * 60 * 60 # signals are available for the previous 3 hours, thus initially start with this time range
        secStart = secInitialStart
        urls.append("")
    urls[0] = time.strftime(fileDateTimeFormat, time.gmtime(secInitialStart)) + " - " + time.strftime(fileDateTimeFormat, time.gmtime(secEnd))

    new = 0
    for sec in range(secStart, secEnd + secStep, secStep):
        t = time.gmtime(sec)
        link = url.format(BO_USER, BO_PASS, BO_STATION)
        link = time.strftime(link, t)
        urls.append(link)
        new += 1
    return urls

def savePendingDownloads(urls):
    global new
    global done
    global dropped
    with open(bookKeepingFile, 'wt') as f:
        for line in urls:
            f.write(line + '\n')
    logger.info("{} downloaded: {} new, {} done, {} dropped, {} pending".format(urls[0], new, done, dropped, len(urls) - 1))

def readStderr(sp, cmd):
    err = ""
    if (sp.returncode):
        exitStatus = {
            "wget": { # see https://www.gnu.org/software/wget/manual/html_node/Exit-Status.html
                0: "No problems occurred.",
                1: "Generic error code.", 
                2: "Parse error - for instance, when parsing command-line options, the '.wgetrc' or '.netrc'...", 
                3: "File I/O error.", 
                4: "Network failure.", 
                5: "SSL verification failure.", 
                6: "Username/password authentication failure.", 
                7: "Protocol errors.", 
                8: "Server issued an error response.",
            },
        }
        if (cmd in exitStatus) and (sp.returncode in exitStatus[cmd]):
            err += "{} returncode: {} ({})".format(cmd, sp.returncode, exitStatus[cmd][sp.returncode])
        else:
            err += "{} returncode: {} (unknown)".format(cmd, sp.returncode)
    line = sp.stderr.readline()
    while line:
        if line.rstrip():
            if "" == err:
                err = "{}: ".format(cmd)
            err += "; {}".format(line.rstrip())
        line = sp.stderr.readline()
    return err

def getStationSignalData(line):
    hasException = False
    signal = {}
    j = {}
    try:
        line = line.rstrip()
        if ',' == line[-1]:
            line = line[0:-1]
        j = json.loads(line)
        if ("station" in j) and (int(BO_STATION) == j["station"]):
            if ("time" in j):
                signal["time_ns"] = j["time"]
                if ("channels" in j):
                    signal["channels"] = 0
                    for c in range(boDb.MAX_STORABLE_NUMBER_OF_CHANNELS):
                        if ("{}".format(c) in j["channels"]):
                            signal["channels"] |= 1 << c # bit mask LSB = channel 0, MSB = hypothetical channel 7
                else:
                    logger.error("no channel in the data line '{}'".format(line))
            else:
                logger.error("no time in the data line '{}'".format(line))
        else:
            logger.error("no 'station' {} in the data line '{}'".format(BO_STATION, line))
    except:
        logger.error("parsing json data failed: '{}'".format(sys.exc_info()[0]))
        hasException = True
    return signal, hasException

def urlGetGrepParse(url):
    result = False
    urlForLogging = url.replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    hasException = False
    signals = []
    signalCnt = 0
    signalInsertFailed = 0
    line = sp1.stdout.readline()
    if not line:
        sp1.wait() # wait for the process to terminate in order to get the returncode
        err = readStderr(sp1, "wget")
        if ("" == err):
            logger.info("'{}' is empty".format(urlForLogging))
        else:
            logger.info("'{}' failed with error '{}'".format(urlForLogging, err))
        return False

    while line:
        signal, hasException = getStationSignalData(line)
        if (signal):
            signals.append(signal)
            signalCnt += 1
        line = sp1.stdout.readline()
    numDone = boDb.insertSignals(signals)
    signalInsertFailed = signalCnt - numDone

    sp1.wait() # wait for the process to terminate in order to get the returncode
    err = readStderr(sp1, "wget")

    if ("" == err):
        loggerText = "'{}' contained {} signals".format(urlForLogging, signalCnt)
        if (signalInsertFailed):
            loggerText += ", insert failed for {}".format(signalInsertFailed)
        if hasException:
            # Exceptions have been observed in the last line. The last line was incomplete.
            # When reading the same file again later, it was error free and contianed additional lines at the end.
            # Looks like at the point in time the data has been read the server is not always ready.
            loggerText += ", json parsing failed -> will retry later"
            result = False
        else:
            result = True
        logger.info(loggerText)
    else:
        logger.debug("'{}' failed with error '{}'".format(urlForLogging, err))
        result = False
    return result

def getBookKeepingPath():
    global bookKeepingFile
    if ("" == BOOK_KEEPING_PATH):
        return
    if (os.path.exists(BOOK_KEEPING_PATH) and os.path.isdir(BOOK_KEEPING_PATH)):
        if ('/' == BOOK_KEEPING_PATH[-1]):
            bookKeepingFile = BOOK_KEEPING_PATH + bookKeepingFile
        else:
            bookKeepingFile = BOOK_KEEPING_PATH + '/' + bookKeepingFile
    else:
        logger.error("BOOK_KEEPING_PATH = '{}' does not exist".format(BOOK_KEEPING_PATH))
        sys.exit(1)

def main():
    # normal program starts here
    global done
    global dropped
    global desiredExecutionTime
    done = 0
    dropped = 0
    configLogger()
    tEnd = time.time() + desiredExecutionTime
    getBookKeepingPath()
    boDb.connectMysql()
    boDb.connectDatabase()
    urlsNow = calcRequiredDownloads()
    urlsFuture = []
    urlsFuture.append(urlsNow.pop(0)) # copy start and end date/time
    time.sleep(30) # allow the server to finalize the file - delay due to the fear to miss something not due to knowledge
    while (tEnd > time.time()):
        if (0 == len(urlsNow)):
            break
        url = urlsNow.pop(0).split(',') # try oldest first
        if (1 == len(url)):
            url.append(0) # no error count present, start counting at 0
        result = urlGetGrepParse(url[0])
        if (True == result):
            done += 1
        else:
            url[1] = int(url[1]) + 1 # increment the error counter
            urlForLogging = url[0].replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
            if (6 <= url[1]): # 1 hour = 60 minutes, best case 1 retry every 10 minutes = 6 retries in 1 hour
                # maximum number of retries is reached
                logger.error("'{}' finally failed after {} attempts".format(urlForLogging, url[1]))
                dropped += 1
            else:
                urlsFuture.append("{},{}".format(url[0], url[1])) # store the incremented error counter
                logger.info("'{}' failed {} times".format(urlForLogging, url[1]))
    for i in range(len(urlsNow)):
        urlsFuture.append("{}".format(urlsNow[i]))
    savePendingDownloads(urlsFuture)
    if done:
        boDb.correlateSignalsStrokes()
    boDb.disconnectMysql()

if __name__ == "__main__":
    main()
